from IPython.display import HTML, Image
from IPython.display import display

from PIL import Image as PILImage
from io import BytesIO

from .uni_loader import concurrent_loads

def peek_df(df, n=1):
    print(df.shape)
    print(df.columns)
    display(df.head(n))

def _src_from_data(data):
    """Base64 encodes image bytes for inclusion in an HTML img element."""
    img_obj = Image(data=data)
    for bundle in img_obj._repr_mimebundle_():
        for mimetype, b64value in bundle.items():
            if mimetype.startswith('image/'):
                return f'data:{mimetype};base64,{b64value}'


def gallery(paths:list[str], labels:list[str] = [], 
            row_height='300px', num_workers=16, debug_print=False, thumbnail_size:int=512):
    """Shows a set of images in a gallery that flexes with the width of the notebook.
    
    Parameters
    ----------
    paths: list of str
        Paths to images to display. Can be local paths, URLs, or S3 paths.

    row_height: str
        CSS height value to assign to all images. Set to 'auto' by default to show images
        with their native dimensions. Set to a value like '250px' to make all rows
        in the gallery equal height.

    num_workers: int
        Number of concurrent workers to load images.

    debug_print: bool
        Whether to print debug information or not.
    
    thumbnail_size: int (optional)
        If provided, resize images to this size using PIL's thumbnail method.
    """
    if len(paths) > 1000:
        raise ValueError("Too many images to display.")
    
    if len(labels) > 0 and len(labels) != len(paths):
        raise ValueError("Number of labels must match number of paths.")
    
    images = concurrent_loads(paths, num_workers=num_workers, debug_print=debug_print)
    
    figures = []
    for i, image in enumerate(images):
        try:
            if thumbnail_size > 0:
                image.thumbnail((thumbnail_size, thumbnail_size))
                
            if image.mode == 'RGBA':
                image = image.convert('RGB')
                
            buffered = BytesIO()
            image.save(buffered, format="JPEG")
            img_data = buffered.getvalue()
            src = _src_from_data(img_data)
            if len(labels) > 0:
                caption_str = labels[i]
            else:
                caption_str = paths[i].split("/")[-1]
            caption = f'<figcaption style="font-size: 0.6em">{caption_str}</figcaption>'
        except Exception as e:
            src = ""
            caption = f'<figcaption style="font-size: 0.6em; color: red;">Error loading {paths[i]}: {e}</figcaption>'
        
        figures.append(f'''
            <figure style="margin: 5px !important;">
              <img src="{src}" style="height: {row_height}">
              {caption}
            </figure>
        ''')
    
    return HTML(data=f'''
        <div style="display: flex; flex-flow: row wrap; text-align: center;">
        {''.join(figures)}
        </div>
    ''')


# Adding the _load_csv and _load_parquet methods to the UniLoader class

from __future__ import annotations

import json
import timeit

import re
import os
import mimetypes
import tempfile
import requests
import orjson

import tomli
import pandas as pd

from PIL import Image
from pathlib import Path
from omegaconf import OmegaConf

from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm.auto import tqdm

from unibox.utils.uni_logger import UniLogger
from unibox.utils.s3_client import S3Client
from unibox.utils.utils import is_url, is_s3_uri


class UniLoader:
    """A simple utility class for loading various file types.
    Data cleaning should be done in the class that uses this loader.
    """

    def __init__(self, logger=None, debug_print=True):
        if not logger:
            self.logger = UniLogger(file_suffix="UniLoader")
        else:
            self.logger = logger

        self.debug_print = debug_print

        # Mapping of file extensions to loader functions
        self.loaders = {
            '.json': self._load_json,
            '.jsonl': self._load_jsonl,  # Added '.jsonl' to the loaders dictionary
            '.txt': self._load_txt,
            '.html': self._load_txt,
            '.csv': self._load_csv,
            '.png': self._load_image,
            '.jpg': self._load_image,
            '.webp': self._load_image,
            '.jpeg': self._load_image,
            '.toml': self._load_toml,
            '.yaml': self._load_yaml,
            '.parquet': self._load_parquet,
            '.feather': self._load_feather,
            ".har": self._load_json,
        }

    def _load_from_s3(self, s3_uri: str):
        """Download a file from an S3 URI and load its content."""
        s3_client = S3Client()
        try:
            with tempfile.TemporaryDirectory() as tmp_dir:
                local_path = Path(s3_client.download(s3_uri, tmp_dir))
                if local_path.exists() and local_path.is_file():
                    return self.loads(local_path)
                else:
                    self.logger.error(f'File was not found at the downloaded path: "{local_path}"')
                    return None
        except Exception as e:
            self.logger.error(f'Error loading from S3 at "{s3_uri}": {e}')
            return None


    def _load_from_url(self, url: str):
        """Download a file from a URL and load its content."""
        response = requests.get(url)
        response.raise_for_status()  # Ensure the request was successful

        _, url_suffix = os.path.splitext(url)
        if url_suffix == '':
            content_type = response.headers.get('content-type')
            mime_suffix = mimetypes.guess_extension(content_type)
            suffix = mime_suffix if mime_suffix else ''
        else:
            suffix = url_suffix

        try:
            with tempfile.NamedTemporaryFile(suffix=suffix, delete=True) as tmp_file:
                tmp_file.write(response.content)
                tmp_file.flush()
                tmp_file_path = tmp_file.name
                # tmp_file.close()
                return self.loads(tmp_file_path)
        except Exception as e:
            self.logger.error(f'Error loading from URL at "{url}": {e}')
            return None

    def loads(self, file_path: Path | str, encoding="utf-8"):
        """Load data from the given file path.

        The type of data returned depends on the file extension.
        """
        start_time = timeit.default_timer()

        # check if is s3 uri or url (downloads the file)
        if isinstance(file_path, str):
            if is_s3_uri(file_path):
                return self._load_from_s3(file_path)

            elif is_url(file_path):
                return self._load_from_url(file_path)

        # is a local file
        if isinstance(file_path, str):
            file_path = Path(file_path)

        file_type = file_path.suffix.lower()

        if not file_path.exists():
            self.logger.warning(f'{file_type} DOES NOT EXIST at "{file_path}"')
            return None

        if file_type not in self.loaders:
            self.logger.error(f'Unsupported file type "{file_type}"')
            return None

        try:
            # read the file using loader function
            result = self.loaders[file_type](file_path, encoding)
            if self.debug_print:
                self.logger.info(f'{file_type} LOADED from "{file_path}" in {timeit.default_timer() - start_time:.2f}s')
            return result
        except Exception as e:
            self.logger.error(f'{file_type} LOAD ERROR at "{file_path}": {e}')
            return None

    def _load_json(self, file_path: Path, encoding='utf-8'):
        try:
            with open(file_path, "rb") as f:
                file_content = f.read()
                # Check if the file content is empty
                if not file_content:
                    self.logger.error(f'{file_path} is empty or zero-length.')
                    return None  # or {} depending on how you want to handle this case
                return orjson.loads(file_content)
        except orjson.JSONDecodeError as e:
            self.logger.error(f'JSON LOAD ERROR at "{file_path}": {e}')
            return None
        except Exception as e:
            self.logger.error(f'Unexpected error loading JSON at "{file_path}": {e}')
            return None


    def _load_jsonl(self, file_path: Path, encoding='utf-8'):
        """
        Load data from a .jsonl file and return it as a list of dictionaries.
        Attempts to handle "NaN" gracefully by replacing them with "null".
        """
        data = []
        with open(file_path, "rb") as f:
            for i, line in enumerate(f, start=1):
                try:
                    # Attempt to decode the line using UTF-8
                    line = line.decode(encoding)
                except UnicodeDecodeError as e:
                    self.logger.error(f"Unicode decode error on line {i}: {e}\nContent: {line}")
                    continue
                try:
                    # Try loading the JSON directly
                    data.append(orjson.loads(line))
                except orjson.JSONDecodeError as initial_error:
                    # If there's a JSON error, check and replace NaN with null
                    if 'NaN' in line:
                        self.logger.warning(f"NaN found and replaced with null on line {i}")
                        line = re.sub(r'\bNaN\b', 'null', line)
                        try:
                            data.append(orjson.loads(line))
                        except orjson.JSONDecodeError as e:
                            self.logger.error(f"JSON decode error on line {i} after replacing NaN: {e}\nContent: {line}")
                    else:
                        self.logger.error(f"JSON decode error on line {i}: {initial_error}\nContent: {line}")
        return data

    def _load_txt(self, file_path: Path, encoding):
        with open(file_path, "r", encoding=encoding) as f:
            raw_lines = f.readlines()
            return [line.strip() for line in raw_lines]

    def _load_csv(self, file_path: Path, encoding):
        return pd.read_csv(file_path, delimiter=',')

    def _load_image(self, file_path: Path, encoding):
        return Image.open(file_path)

    def _load_toml(self, file_path: Path, encoding):
        with open(file_path, "r", encoding=encoding) as f:
            return tomli.load(f)

    def _load_yaml(self, file_path: Path, encoding):
        with open(file_path, "r", encoding=encoding) as f:
            return OmegaConf.load(f)

    def _load_parquet(self, file_path: Path, encoding):
        return pd.read_parquet(file_path)

    def _load_feather(self, file_path: Path, encoding):
        return pd.read_feather(file_path)


def concurrent_loads(uris_list, num_workers=8, debug_print=True):
    """
    Loads dataframes concurrently from a list of S3 URIs.

    :param uris_list: list of S3 URIs (or local) to load
    :param num_workers: int, number of concurrent workers
    :param debug_print: bool, whether to print debug information or not
    :return: list of loaded dataframes

    >>> selected_uris = [f"{base_s3_uri}/{i}.merged.parquet" for i in selected_ids]
    >>> dfs = concurrent_loads(selected_uris, num_workers)
    >>> df = pd.concat(dfs, ignore_index=True)
    """
    loader = UniLoader(debug_print=False)

    results = []
    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        future_to_uri = {executor.submit(
            loader.loads, curr_uri): curr_uri for curr_uri in uris_list}

        if debug_print:
            futures_iter = tqdm(as_completed(future_to_uri), total=len(
                uris_list), desc="Loading batches", mininterval=3)
        else:
            futures_iter = as_completed(future_to_uri)


        for future in futures_iter:
            curr_uri = future_to_uri[future]
            try:
                _res = future.result()
                results.append(_res)
            except Exception as e:
                print(f"Exception for {curr_uri}: {e}")
    return results


def sample_usage():
    # Usage example
    logger = UniLogger("logs", file_suffix="data_loader")
    data_loader = UniLoader(logger)
    json_data = data_loader.loads("example.json")  # string
    txt_data = data_loader.loads(Path("example.txt"))  # path
    csv_data = data_loader.loads(Path("example.csv"))
    image_data = data_loader.loads(Path("example.png"))
    toml_data = data_loader.loads(Path("example.toml"))
    yaml_data = data_loader.loads(Path("example.yaml"))
    parquet_data = data_loader.loads(Path("example.parquet"))


def debug_url():
    loader = UniLoader()
    img = loader.loads("https://cdn.donmai.us/sample/8e/ea/__karyl_princess_connect_drawn_by_maxwelzy__sample-8eea944690c0c0b27e303420cb1e65bd.jpg")

    print(type(img))


def debug_jsonl():
    loader = UniLoader()
    # /rmtrain/quail/sdxl_qft/sdxl_qft_v5.jsonl
    data = loader.loads("/home/ubuntu/dev/unibox/noteooks/sdxl_qft_v5.jsonl")
    print(data)


if __name__ == "__main__":
    debug_jsonl()




import os
import json
import math
import pandas as pd
from PIL import Image, JpegImagePlugin, PngImagePlugin
from typing import Union

from unibox.utils.uni_logger import UniLogger
from unibox.utils.s3_client import S3Client  # Assuming S3Client is in a module named s3_client


class UniSaver:
    """A simple utility class for saving various data types to appropriate file formats."""
    
    def __init__(self, logger=None, debug_print=True, s3_client=None):
        self.logger = logger or UniLogger(file_suffix="UniSaver")
        self.debug_print = debug_print
        self.s3_client = s3_client or S3Client()

    def saves(self, data, file_path: str):
        """Save data to the given file path."""
        data_type = type(data).__name__
        expected_extension = self._get_expected_extension(data_type, data)
        file_path = self._validate_and_append_extension(file_path, expected_extension)

        is_s3 = file_path.startswith("s3://")
        local_file_path = self._get_local_file_path(file_path, is_s3)

        try:
            self._save_data(data, data_type, local_file_path, expected_extension)
            if is_s3:
                self._upload_to_s3(local_file_path, file_path)
            if self.debug_print:
                self.logger.info(f'{data_type} saved successfully to "{file_path}"')
        
        except PermissionError:
            self._handle_permission_error(data, data_type, local_file_path, expected_extension)
            if self.debug_print:
                self.logger.info(f'{data_type} saved successfully to "{file_path}"')
        
        except Exception as e:
            self.logger.error(f'{data_type} save ERROR at "{local_file_path}": {e}')

    def _get_expected_extension(self, data_type, data):
        if data_type == 'list':
            extension = self._list_extension(data)
        else:
            extension_mapping = {
                'dict': '.json',
                'DataFrame': '.parquet',
                'str': '.txt'
            }

            # Generalize for Image types
            if isinstance(data, Image.Image):
                extension = '.png'
            elif data_type not in extension_mapping:
                self.logger.error(f'Unsupported data type "{data_type}"')
                raise ValueError(f'Unsupported data type "{data_type}"')
            else:
                extension = extension_mapping[data_type]
        
        return extension

    def _validate_and_append_extension(self, file_path: str, expected_extension: str) -> str:
        if not file_path.endswith(expected_extension):
            if '.' in file_path:
                self.logger.error(f'Invalid file extension for expected "{expected_extension}" but got "{file_path.split(".")[-1]}"')
                self.logger.warning("File extension will be appended to the file name")

            file_path += expected_extension
            self.logger.warning(f"File without extension, saving to actual path: {file_path}")
        
        return file_path

    def _get_local_file_path(self, file_path: str, is_s3: bool) -> str:
        if is_s3:
            return os.path.basename(file_path)  # Use only the basename for the local file path
        return file_path

    def _save_data(self, data, data_type, file_path, expected_extension=None):
        
        if isinstance(data, Image.Image): # Handle image data separately
            self._save_image(data, file_path)
            return
        
        save_function_mapping = {
            'dict': self._save_json,
            'list': lambda d, p: self._save_list(d, p, expected_extension),
            'set': lambda d, p: self._save_list(list(d), p, expected_extension),
            'DataFrame': self._save_parquet,
            'str': lambda d, p: self._save_txt([d], p)
        }

        if data_type not in save_function_mapping:
            self.logger.error(f'No save function available for data type "{data_type}"')
            raise ValueError(f'No save function available for data type "{data_type}"')

        save_function_mapping[data_type](data, file_path)

    def _upload_to_s3(self, local_file_path: str, s3_file_path: str):
        s3_dir = os.path.dirname(s3_file_path)
        self.s3_client.upload(local_file_path, s3_dir)
        os.remove(local_file_path)

    def _handle_permission_error(self, data, data_type, local_file_path, expected_extension):
        alternative_file_path = local_file_path.replace('.parquet', '_alternative.parquet')
        self.logger.warning(f'Permission denied for "{local_file_path}". Trying to save to "{alternative_file_path}" instead.')
        self._save_data(data, data_type, alternative_file_path, expected_extension)

    def _list_extension(self, data: list) -> str:
        if all(isinstance(item, dict) for item in data):
            return '.jsonl'
        elif all(isinstance(item, str) for item in data):
            return '.txt'
        else:
            raise TypeError("List content must be either all dictionaries or all strings.")

    def _replace_special_values(self, obj):
        if isinstance(obj, dict):
            return {k: (None if isinstance(v, float) and math.isnan(v) else v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [None if isinstance(v, float) and math.isnan(v) else v for v in obj]
        return obj

    def _save_json(self, data: dict, file_path: str):
        data = self._replace_special_values(data)
        with open(file_path, "w", encoding="utf-8") as f:            
            json.dump(data, f, ensure_ascii=False, indent=2)

    def _save_list(self, data: list, file_path: str, extension: str):
        if extension == '.jsonl':
            self._save_jsonl(data, file_path)
        elif extension == '.txt':
            self._save_txt(data, file_path)
        else:
            raise ValueError("Invalid extension for list data.")

    def _save_jsonl(self, data: list, file_path: str):
        with open(file_path, "w", encoding="utf-8") as f:
            for item in data:
                item = self._replace_special_values(item)
                json.dump(item, f, ensure_ascii=False)
                f.write('\n')

    def _save_txt(self, data: list, file_path: str):
        with open(file_path, "w", encoding="utf-8") as f:
            for item in data:
                f.write(f'{item}\n')

    def _save_parquet(self, data: pd.DataFrame, file_path: str):
        data.to_parquet(file_path, index=False)

    def _save_image(self, data: Image.Image, file_path: str):
        if not isinstance(data, (JpegImagePlugin.JpegImageFile, PngImagePlugin.PngImageFile)):
            data = data.convert("RGB")  # Convert to standard Image object if not already one
        data.save(file_path)


def sample_usage():
    # Usage example
    logger = UniLogger("logs", file_suffix="saver")
    saver = UniSaver(logger=logger, s3_client=s3_client)
    
    # Save to local
    json_data = {'key': 'value'}
    saver.saves(json_data, "example")  # Automatically adds .json
    jsonl_data = [{'key1': 'value1'}, {'key2': 'value2'}]
    saver.saves(jsonl_data, "example.jsonl")
    txt_data = ['line1', 'line2']
    saver.saves(txt_data, "example.txt")
    image_data = Image.new('RGB', (60, 30), color='red')
    saver.saves(image_data, "example.png")
    df_data = pd.DataFrame({'A': [1, 2], 'B': [3, 4]})
    saver.saves(df_data, "example.parquet")
    
    # Save to S3
    saver.saves(json_data, "s3://my-bucket/example.json")
    saver.saves(image_data, "s3://my-bucket/example.png")
    saver.saves(df_data, "s3://my-bucket/example.parquet")


def debug_saver():
    # import pandas as pd
    # sample_df = pd.DataFrame([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    # saver = UniSaver()
    # saver.saves(sample_df, "s3://unidataset-danbooru/sample_df.parquet")

    from PIL import Image
    from unibox.utils.uni_loader import UniLoader

    loader = UniLoader()
    jpeg_file = loader.loads("https://cdn.donmai.us/180x180/8e/ea/8eea944690c0c0b27e303420cb1e65bd.jpg")

    
    saver = UniSaver()
    saver.saves(jpeg_file, "s3://unidataset-danbooru/sample_image.jpg")
